#!/usr/bin/env python
# -*- coding: utf-8 -*-


# ==============================================================================
#
#       File Name : letsagree/querysets.py
#
#       Creation Date : Tue 29 Jan 2019 10:51:29 AM EET (10:51)
#
#       Last Modified : Fri 22 Mar 2019 11:48:45 PM EET (23:48)
#
# ==============================================================================

from django.db import models
from django.db.models import F, Window
from django.db.models.functions import FirstValue


class TermQS(models.QuerySet):
    def prepare_active_terms(self, user_id):
        """
        The query opens a window, groups all entries by the group they belong to,
        orders them by the date and annotates the first one of each group
        to a field called 'active_terms'.
        """
        return (
            self.filter(group_key__user=user_id)
            .annotate(
                active_terms=Window(
                    expression=FirstValue("id"),
                    partition_by=["group_key"],
                    order_by=F("date_created").desc(),
                )
            )
            .distinct()
        )

    def get_pending_terms(self, user_id):
        """
        Query that returns the user's pending to be signed terms.

        1 db hit.

        values_list('active_terms') returns only the latest active term
        which is currently in effect for each group.

        Consequently, it is possible for a user to miss -and never agree with-
        a term that was in effect in a period during which he did not
        happen to log in. He will agree, however, to the newest term that
        follows the missed one.
        """
        return self.filter(
            id__in=self.prepare_active_terms(user_id).values_list("active_terms")
        ).exclude(users_agreed__user_key_id=user_id)

    def get_signed_agreements(self, user_id):
        """
        Query that returns the user's signed agreements.

        2 db hits

        This query prefetches, for each term, the agreement made by the user,
        in order to allow access without hitting the database when
        the date of the signature has to be retrieved. One agreement per
        term is expected, hoewever and order_by is provided in case more
        agreements exist.

        values_list('active_terms') returns only the latest term
        which is currently in effect for each group.
        """
        from letsagree.models import NotaryPublic

        return (
            self.filter(
                id__in=self.prepare_active_terms(user_id).values_list("active_terms")
            )
            .filter(users_agreed__user_key_id=user_id)
            .prefetch_related(
                models.Prefetch(
                    "users_agreed",
                    queryset=(
                        NotaryPublic.objects.filter(user_key_id=user_id).order_by(
                            "-date_signed"
                        )
                    ),
                    to_attr="signature_dates",
                )
            )
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-


# ==============================================================================
#
#       File Name : letsagree/views.py
#
#       Creation Date : Sun 27 Jan 2019 07:54:42 PM EET (19:54)
#
#       Last Modified : Tue 07 Apr 2020 11:43:40 PM EEST (23:43)
#
# ==============================================================================

from django.db import transaction
from django.conf import settings
from django.core.cache import cache
from django.http import Http404
from django.urls import reverse, NoReverseMatch
from django.utils.translation import gettext_lazy as _
from django.views.generic import FormView
from letsagree import models
from letsagree.forms import PendingAgreementFormSet


class PendingView(FormView):
    http_method_names = ["get", "post"]
    template_name = "letsagree/pending.html"
    form_class = PendingAgreementFormSet

    def dispatch(self, request, *args, **kwargs):
        if not request.user.is_authenticated:
            raise Http404()
        self.success_url = request.GET.get("next")
        return super().dispatch(request, *args, **kwargs)

    def get_form_kwargs(self):
        """
        Pass to the modelformset_factory a queryset argument to create the
        formset that represents the terms needing the user's consent.
        """
        kwargs = super().get_form_kwargs()
        # Avoid KeyError that randomly occurs
        user_id = self.request.user.id
        kwargs["queryset"] = models.Term.objects.get_pending_terms(user_id)
        return kwargs

    def form_valid(self, form):
        """
        The user has agreed to the terms, save each agreement in the database.

        bulk_create could be used, but it is only compatible with PostgreSQL
        which, at the moment, is the only db able to handle autoincremented pk.
        """
        user_id = self.request.user.id
        for sub_form in form:
            with transaction.atomic():
                models.NotaryPublic.objects.create(
                    term_key_id=sub_form.instance.id, user_key_id=user_id
                )
        cache_key = "letsagree-{0}".format(user_id)
        cache.set(cache_key, False, 24 * 3600)
        return super().form_valid(form)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["browser_title"] = getattr(
            settings, "LETSAGREE_BROWSER_TITLE", _("Let's Agree")
        )
        context["border_header"] = getattr(settings, "LETSAGREE_BORDER_HEADER", "")
        context["user"] = self.request.user

        try:
            logout_url_app = (
                getattr(settings, "LETSAGREE_LOGOUT_APP_NAME", "admin") or "admin"
            )
            context["logout_url"] = reverse("{0}:logout".format(logout_url_app))
        except NoReverseMatch:
            pass

        if len(context["form"]) == 0:
            context["empty_form"] = True
        else:
            context["empty_form"] = False

        return context



#!/usr/bin/env python


# ==============================================================================
#
#       File Name : letsagree/middleware.py
#
#       Creation Date : Mon 28 Jan 2019 01:20:20 PM EET (13:20)
#
#       Last Modified : Fri 27 Mar 2020 08:50:03 PM EET (20:50)
#
# ==============================================================================

from django.conf import settings
from django.core.cache import cache
from django.shortcuts import redirect
from django.urls import reverse
from letsagree.models import Term


class LetsAgreeMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        redirect_url = ConsentEvaluator(request).get_redirect_url()
        if redirect_url:
            return redirect(redirect_url)
        else:
            return self.get_response(request)


class ConsentEvaluator:
    def __init__(self, request):
        self.user_id = request.user.id
        self.path = request.path
        self.request_needs_investigation = self.validate_user_intention()
        self.get_next = request.GET.get("next") or request.path

    def validate_user_intention(self):
        """
        Check if a logged user exists and whether this user is requesting a url
        that requires previous consent of the relevant tos.
        excluded urls form consent are:
            * The letsagree form consent url
            * The logout url
            * All letsagree admin urls, except for the request to add a new tos.
              (the user who is entitled to add a new tos, has only the right to
              do so if he has already agreed to the current term in effect.)
            *
        """
        logout_url_app = (
            getattr(settings, "LETSAGREE_LOGOUT_APP_NAME", "admin") or "admin"
        )
        logout_url = (
            reverse("{0}:logout".format(logout_url_app)) if logout_url_app else ""
        )
        satisfied_prerequisites = all(
            (
                self.user_id,
                self.path != reverse("letsagree:pending"),
                self.path != logout_url,
                any(
                    (
                        not self.path.startswith(
                            reverse("admin:app_list", args=("letsagree",))
                        ),
                        "add" in self.path,
                    )
                ),
            )
        )
        return satisfied_prerequisites

    def get_or_set_cache(self, cache_key):
        """
        if cache exists:
            return its value
        else:
            set cache and return its new value
        """
        user_consent_required = cache.get(cache_key, None)

        if user_consent_required is None:
            cache_value = Term.objects.get_pending_terms(self.user_id).exists()
            cache.set(cache_key, cache_value, 24 * 3600)
            return cache_value
        else:
            return user_consent_required

    def consent_is_required(self):
        """
        Return True if consent is required, else False.
        If cache is enabled:
            * check or set the cache
        else;
            * query for the status

        By default cache is deactivated because it exposes the user id which
        uniquely identifies a user in the database. Nonetheless,
        django-hashid-field (https://github.com/nshafer/django-hashid-field),
        can obscure the user id while keeping its uniqueness.
        """
        if getattr(settings, "LETSAGREE_CACHE", False):
            cache_key = "letsagree-{0}".format(self.user_id)
            return self.get_or_set_cache(cache_key)

        else:
            return Term.objects.get_pending_terms(self.user_id).exists()

    def get_next_parameter(self):
        """
            If next parameter exists in request, set it also in the redirect
            unless it equals the url of the consent form.
        """
        return (
            "?next={0}".format(self.get_next)
            if self.get_next and self.get_next != reverse("letsagree:pending")
            else None
        )

    def get_redirect_url(self):
        """
        If user has already agreed:
            * Return None
        else:
            * Return Redirect URL
        """
        result = None
        if self.request_needs_investigation:
            if self.consent_is_required():
                redirect_url = reverse("letsagree:pending")
                next_url = self.get_next_parameter()
                result = "{0}{1}".format(redirect_url, next_url)
        return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-


# ==============================================================================
#
#       File Name : letsagree/forms.py
#
#       Creation Date : Mon 25 Feb 2019 05:44:28 PM EET (17:44)
#
#       Last Modified : Mon 08 Apr 2019 03:16:38 PM EEST (15:16)
#
# ==============================================================================

from django import forms
from django.conf import settings
from django.utils.translation import gettext_lazy as _
from letsagree import models
from translated_fields import to_attribute


class PendingConsentForm(forms.ModelForm):
    agree = forms.BooleanField(required=True, label=_("I Give my Consent"))

    def save(self, commit=False):
        """
        There is nothing to save. This form is a building block of a formset
        with read-only contents.
        """
        pass

    def __init__(self, *args, **kwargs):
        """
        All Fields are disabled. They are rendered as read-only fields.
        """
        super().__init__(*args, **kwargs)
        for field in self.fields.keys():
            if field == "agree":
                continue
            self.fields[field].disabled = True

    class Media:
        css = getattr(settings, "LETSAGREE_CSS", dict())
        js = getattr(settings, "LETSAGREE_JS", tuple())

    class Meta:
        model = models.Term
        fields = (
            "date_created",
            to_attribute("summary"),
            to_attribute("content"),
            "agree",
        )


PendingAgreementFormSet = forms.modelformset_factory(
    models.Term, form=PendingConsentForm, extra=0
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-


# ==============================================================================
#
#       File Name : letsagree/models.py
#
#       Creation Date : Sun 27 Jan 2019 07:54:42 PM EET (19:54)
#
#       Last Modified : Mon 01 Apr 2019 06:02:32 PM EEST (18:02)
#
# ==============================================================================

from django.db import models
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from letsagree.querysets import TermQS
from translated_fields import TranslatedFieldWithFallback


class BaseRelatedManager(models.Manager):
    def get_queryset(self):
        return super().get_queryset().select_related()


class Term(models.Model):
    group_key = models.ForeignKey(
        Group,
        on_delete=models.PROTECT,
        related_name="terms",
        verbose_name=_("Related Group"),
    )
    title = TranslatedFieldWithFallback(
        models.CharField(max_length=63, verbose_name=_("Title"), db_index=True)
    )
    summary = TranslatedFieldWithFallback(models.TextField(verbose_name=_("Summary")))
    content = TranslatedFieldWithFallback(
        models.TextField(verbose_name=_("Terms and Conditions"))
    )

    date_created = models.DateTimeField(
        verbose_name=_("Date and Time of Document Creation"),
        default=timezone.now,
        db_index=True,
    )

    objects = BaseRelatedManager.from_queryset(TermQS)()

    def __str__(self):
        return '{0}, Group:"{1}" on {2}, {3}'.format(
            self.id,
            self.group_key.name,
            self.date_created.strftime("%Y-%m-%d-%T"),
            self.title,
        )

    class Meta:
        verbose_name = _("Terms & Conditions")
        verbose_name_plural = _("Terms & Conditions")


class NotaryPublic(models.Model):
    user_key = models.ForeignKey(
        get_user_model(),
        verbose_name=_("User"),
        on_delete=models.PROTECT,
        related_name="agreed_terms",
    )
    term_key = models.ForeignKey(
        Term,
        verbose_name=_("Terms and Conditions"),
        on_delete=models.PROTECT,
        related_name="users_agreed",
    )
    date_signed = models.DateTimeField(
        verbose_name=_("Date and Time of User Consent"),
        default=timezone.now,
        db_index=True,
    )

    objects = BaseRelatedManager()

    def __str__(self):
        return "{0}: User:{1}, Term-id:{2}".format(
            self.id, self.user_key.username, self.term_key_id
        )

    class Meta:
        verbose_name = _("Notary Public")
        verbose_name_plural = _("Notary Public")
        unique_together = ("user_key", "term_key")



#!/usr/bin/env python
# -*- coding: utf-8 -*-


# ==============================================================================
#
#       File Name : letsagree/admin.py
#
#       Creation Date : Sun 27 Jan 2019 07:54:42 PM EET (19:54)
#
#       Last Modified : Sun 20 Oct 2019 02:31:39 PM EEST (14:31)
#
# ==============================================================================

from django.conf import settings
from django.contrib import admin
from letsagree import models
from translated_fields import TranslatedFieldAdmin, to_attribute
from django.http import HttpResponseRedirect
from django.urls import reverse


term_parents = (
    (admin.ModelAdmin,)
    if len(settings.LANGUAGES) < 2
    else (TranslatedFieldAdmin, admin.ModelAdmin)
)


@admin.register(models.Term)
class TermAdmin(*term_parents):
    autocomplete_fields = ("group_key",)
    list_display = ("id", "group_key", "date_created", "title")
    list_display_links = ("id", "date_created", "title")
    readonly_fields = ("date_created",)
    search_fields = ("id", to_attribute("title"))

    @staticmethod
    def has_delete_permission(request, obj=None):
        return False

    @staticmethod
    def has_change_permission(request, obj=None):
        return False

    def get_queryset(self, request):
        qs = super().get_queryset(request)
        if request.user.has_perm("letsagree.add_term"):
            return qs
        else:
            return qs.filter(group_key__user=request.user)


@admin.register(models.NotaryPublic)
class NotaryPublicAdmin(admin.ModelAdmin):
    autocomplete_fields = ("term_key", "user_key")
    list_display = ("id", "date_signed", "user_key", "term_key")
    list_display_links = ("id", "date_signed", "term_key")
    readonly_fields = ("date_signed",)
    search_fields = ("id", "user_key__username")

    def get_queryset(self, request):
        qs = super().get_queryset(request)
        if request.user.is_superuser:
            return qs
        return qs.filter(user_key_id=request.user.id)

    @staticmethod
    def has_add_permission(request):
        return False

    @staticmethod
    def has_delete_permission(request, obj=None):
        if obj:
            return request.user.id == obj.user_key_id
        return False

    @staticmethod
    def has_change_permission(request, obj=None):
        return False

    def response_delete(self, request, obj_display, obj_id):
        return HttpResponseRedirect(reverse("admin:letsagree_notarypublic_changelist"))



#!/usr/bin/env python
# -*- coding: utf-8 -*-


# ==============================================================================
#
#       File Name : test_setup/i18n_settings.py
#
#       Creation Date : Tue 09 Apr 2019 01:19:13 AM EEST (01:19)
#
#       Last Modified : Mon 15 Apr 2019 07:57:12 PM EEST (19:57)
#
# ==============================================================================

import os
from collections import namedtuple

Settings = namedtuple("Settings", ["username", "password"])

db = {
    "django.db.backends.postgresql": Settings(
        username=os.environ.get("POSTGRES_USER", os.environ.get("TOX_DB_USER")),
        password=os.environ.get("POSTGRES_PASSWD", os.environ.get("TOX_DB_PASSWD")),
    ),
    "django.db.backends.mysql": Settings(
        username=os.environ.get("MARIADB_USER", os.environ.get("TOX_DB_USER")),
        password=os.environ.get("MARIADB_PASSWD", os.environ.get("TOX_DB_PASSWD")),
    ),
}

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

DATABASES = {
    "default": {
        "ENGINE": os.environ["TOX_DB_ENGINE"],
        "NAME": os.environ["TOX_DB_NAME"],
        "USER": db[os.environ["TOX_DB_ENGINE"]].username,
        "PASSWORD": db[os.environ["TOX_DB_ENGINE"]].password,
    }
}


INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "letsagree",
]

LANGUAGE_CODE = "fr"

LANGUAGES = (("fr", "French"), ("en", "English"))

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.locale.LocaleMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
    "letsagree.middleware.LetsAgreeMiddleware",
]

ROOT_URLCONF = "test_setup.urls"

SECRET_KEY = "fsfie434*^%Dkkvkdnf8(*^@#()Fjvhdfi3))^%$"

STATIC_URL = "/static/"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
                "django.template.context_processors.i18n",
            ]
        },
    }
]

USE_I18N = True

USE_L10N = True
